# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import enum
from http import client as http_client
import io
import json
from unittest import mock
import zipfile

from sushy import exceptions
from sushy.resources import base as resource_base
from sushy.tests.unit import base


BASE_RESOURCE_JSON = {
    "@odata.type": "#FauxResource.v1_0_0.FauxResource",
    "Id": "1111AAAA",
    "Name": "Faux Resource",
    "@odata.id": "/redfish/v1/FauxResource/1111AAAA",
    "Oem": {
        "Contoso": {
            "@odata.type": "http://contoso.com/schemas/extensions.v1_2_1#contoso.AnvilTypes1",  # noqa
            "slogan": "Contoso never fail",
            "disclaimer": "* Most of the time"
        },
        "EID_412_ASB_123": {
            "@odata.type": "http://AnotherStandardsBody/schemas.v1_0_1#styleInfoExt",  # noqa
            "Style": "Executive"
        }
    },
    "Links": {
        "Oem": {
            "Contoso": {
                "@odata.type": "http://contoso.com/schemas/extensions.v1_2_1#contoso.AnvilTypes1",  # noqa
                "slogan": "Contoso never fail",
                "disclaimer": "* Most of the time"
            },
            "EID_420_ASB_345": {
                "@odata.type": "http://AnotherStandardsBody/schemas.v1_0_1#styleInfoExt",  # noqa
                "Style": "Executive"
            }
        }
    }

}


class BaseResource(resource_base.ResourceBase):

    def _parse_attributes(self, json_doc):
        pass


class BaseResource2(resource_base.ResourceBase):
    pass


class ResourceBaseTestCase(base.TestCase):

    def setUp(self):
        super(ResourceBaseTestCase, self).setUp()
        self.conn = mock.Mock()
        self.conn.get.return_value.json.return_value = (
            copy.deepcopy(BASE_RESOURCE_JSON))
        self.base_resource = BaseResource(connector=self.conn, path='/Foo',
                                          redfish_version='1.0.2')
        self.assertFalse(self.base_resource._is_stale)
        self.base_resource2 = BaseResource2(connector=self.conn, path='/Foo',
                                            redfish_version='1.0.2')
        # refresh() is called in the constructor
        self.conn.reset_mock()

    def test_refresh_no_force(self):
        self.base_resource.refresh(force=False)
        self.conn.get.assert_not_called()

    def test_refresh_force(self):
        self.base_resource.refresh()
        self.conn.get.assert_called_once_with(path='/Foo')

    def test_invalidate(self):
        self.base_resource.invalidate()
        self.conn.get.assert_not_called()

        self.base_resource.refresh(force=False)
        self.conn.get.assert_called_once_with(path='/Foo')

    def test_invalidate_force_refresh(self):
        self.base_resource.invalidate(force_refresh=True)
        self.conn.get.assert_called_once_with(path='/Foo')

    def test_refresh_archive(self):
        mock_response = mock.Mock(
            headers={'content-type': 'application/zip'})
        with open('sushy/tests/unit/json_samples/TestRegistry.zip', 'rb') as f:
            mock_response.content = f.read()
        self.conn.get.return_value = mock_response

        resource = BaseResource(connector=self.conn,
                                path='/Foo',
                                redfish_version='1.0.2',
                                reader=resource_base.
                                JsonArchiveReader('Test.2.0.json'))

        self.assertIsNotNone(resource._json)
        self.assertEqual('Test.2.0.0', resource._json['Id'])

    @mock.patch.object(resource_base, 'LOG', autospec=True)
    def test_refresh_archive_not_implemented(self, mock_log):
        mock_response = mock.Mock(
            headers={'content-type': 'application/gzip'})
        self.conn.get.return_value = mock_response
        BaseResource(connector=self.conn,
                     path='/Foo',
                     redfish_version='1.0.2',
                     reader=resource_base.JsonArchiveReader('Test.2.0.json'))
        mock_log.error.assert_called_once()

    @mock.patch.object(io, 'BytesIO', autospec=True)
    def test_refresh_archive_badzip_error(self, mock_io):
        mock_response = mock.Mock(
            headers={'content-type': 'application/zip'})
        mock_io.side_effect = zipfile.BadZipfile('Something wrong')
        self.conn.get.return_value = mock_response

        self.assertRaises(exceptions.SushyError,
                          BaseResource, connector=self.conn,
                          path='/Foo',
                          redfish_version='1.0.2',
                          reader=resource_base.
                          JsonArchiveReader('Test.2.0.json'))

    def test_init_default_reader(self):
        resource_a = BaseResource(connector=self.conn)
        resource_b = BaseResource(connector=self.conn)

        self.assertIsInstance(resource_a._reader, resource_base.JsonDataReader)
        self.assertIsInstance(resource_b._reader, resource_base.JsonDataReader)

        self.assertIsNot(resource_a._reader, resource_b._reader)

    def test__parse_attributes(self):
        expected_oem_vendors = ['Contoso', 'EID_412_ASB_123',
                                'EID_420_ASB_345']
        actual_oem_vendors = sorted(self.base_resource2.oem_vendors)
        self.assertEqual(expected_oem_vendors, actual_oem_vendors)
        self.assertEqual('base_resource2', self.base_resource2.resource_name)

    def test_refresh_local(self):
        resource = BaseResource(None, 'json_samples/message_registry.json',
                                reader=resource_base.
                                JsonPackagedFileReader('sushy.tests.unit'))
        self.assertIsNotNone(resource._json)
        self.assertEqual('Test.1.1.1', resource._json['Id'])

    def test_refresh_public(self):
        mock_connector = mock.Mock()
        with open('sushy/tests/unit/json_samples/message_registry.json') as f:
            mock_connector.get.return_value.json.return_value = json.load(f)
        resource = BaseResource(mock_connector, 'https://example.com/'
                                'message_registry.json',
                                reader=resource_base.JsonPublicFileReader())
        mock_connector.get.assert_called_once_with('https://example.com/'
                                                   'message_registry.json')
        self.assertIsNotNone(resource._json)
        self.assertEqual('Test.1.1.1', resource._json['Id'])


class TestResource(resource_base.ResourceBase):
    """A concrete Test Resource to test against"""

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None, root=None):
        """Ctor of TestResource

        :param connector: A Connector instance
        :param identity: The id of the Resource
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages.
        """
        super(TestResource, self).__init__(connector, 'Fakes/%s' % identity,
                                           redfish_version, registries, root)
        self.identity = identity

    def _parse_attributes(self, json_doc):
        pass


class TestResourceCollection(resource_base.ResourceCollectionBase):
    """A concrete Test Resource Collection to test against"""

    @property
    def _resource_type(self):
        return TestResource

    def __init__(self, connector, redfish_version=None, registries=None,
                 root=None):
        """Ctor of TestResourceCollection

        :param connector: A Connector instance
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages.
        """
        super(TestResourceCollection, self).__init__(
            connector, 'Fakes', redfish_version, registries, root)


class ResourceCollectionBaseTestCase(base.TestCase):

    def setUp(self):
        super(ResourceCollectionBaseTestCase, self).setUp()
        self.conn = mock.MagicMock()
        self.test_resource_collection = TestResourceCollection(
            self.conn, redfish_version='1.0.x', registries=None)
        self.conn.reset_mock()

    def test_get_member(self):
        # | GIVEN |
        # setting a valid member identity
        self.test_resource_collection.members_identities = ('1',)
        # | WHEN |
        result = self.test_resource_collection.get_member('1')
        # | THEN |
        self.assertIsInstance(result, TestResource)
        self.assertEqual('1', result.identity)
        self.assertEqual('1.0.x', result.redfish_version)

    def test_get_member_for_invalid_id(self):
        # | GIVEN |
        # setting a valid member identity
        self.test_resource_collection.members_identities = ('1',)
        self.conn.get.side_effect = exceptions.ResourceNotFoundError(
            method='GET', url='http://foo.bar:8000/redfish/v1/Fakes/2',
            response=mock.MagicMock(status_code=http_client.NOT_FOUND))
        # | WHEN & THEN |
        self.assertRaises(exceptions.ResourceNotFoundError,
                          self.test_resource_collection.get_member, '2')
        self.conn.get.assert_called_once_with(path='Fakes/2')

    def _validate_get_members_result(self, member_ids):
        # | GIVEN |
        # setting some valid member paths
        self.test_resource_collection.members_identities = member_ids
        # | WHEN |
        result = self.test_resource_collection.get_members()
        # | THEN |
        self.assertIsInstance(result, list)
        for val in result:
            self.assertIsInstance(val, TestResource)
            self.assertIn(val.identity, member_ids)
            self.assertEqual('1.0.x', val.redfish_version)
            self.assertFalse(val._is_stale)

        return result

    def test_get_members(self):
        self._validate_get_members_result(('1', '2'))

    def test_get_members_on_refresh(self):
        all_members = self._validate_get_members_result(('1', '2'))

        # Call resource invalidate
        self.test_resource_collection.invalidate()
        self.assertTrue(self.test_resource_collection._is_stale)
        #  Now invoke refresh action on resource. This can be viewed as
        # "light refresh" which involves only the resource's fresh retrieval
        # and not its nested resources (these are only marked as stale).
        self.test_resource_collection.refresh(force=False)
        # resource itself is fresh
        self.assertFalse(self.test_resource_collection._is_stale)
        # members are marked as stale
        for m in all_members:
            self.assertTrue(m._is_stale)

        self._validate_get_members_result(('1', '2'))
        # members are also now freshly retrieved
        for m in all_members:
            self.assertFalse(m._is_stale)

        # Again invalidate and do a forced refresh on resource
        self.test_resource_collection.invalidate(force_refresh=True)
        # Now, even the members are also freshly retrieved. This can be viewed
        # as "cascading refresh" which involves not only the resource's fresh
        # retrieval but also its nested resources.
        for m in all_members:
            self.assertFalse(m._is_stale)

    def test_get_members_caching(self):
        result = self._validate_get_members_result(('1', '2'))
        self.assertIs(result, self.test_resource_collection.get_members())


TEST_JSON = {
    'String': 'a string',
    'Integer': '42',
    'MappedList': ['raw1', 'raw2', 'raw'],
    'Nested': {
        'String': 'another string',
        'Integer': 0,
        'Object': {
            'Field': 'field value'
        },
        'Mapped': 'raw'
    },
    'ListField': [
        {
            'String': 'a third string',
            'Integer': 1
        },
        {
            'String': 'a fourth string',
            'Integer': 2
        }
    ],
    'Dictionary': {
        'key1': {'property_a': 'value1', 'property_b': 'value2'},
        'key2': {'property_a': 'value3', 'property_b': 'value4'}
    },
    'Enum': 'PROTOCOL_FIELD_2',
    'EnumList': ['PROTOCOL_FIELD_2', 'PROTOCOL_FIELD_3'],
}


MAPPING = {
    'raw': 'real',
    'raw1': 'real1',
    'raw2': 'real2'
}


class EnumMapping(enum.Enum):

    FIELD1 = "PROTOCOL_FIELD_1"
    FIELD2 = "PROTOCOL_FIELD_2"
    FIELD3 = "PROTOCOL_FIELD_3"


class NestedTestField(resource_base.CompositeField):
    string = resource_base.Field('String', required=True)
    integer = resource_base.Field('Integer', adapter=int)
    nested_field = resource_base.Field(['Object', 'Field'], required=True)
    mapped = resource_base.MappedField('Mapped', MAPPING)
    non_existing = resource_base.Field('NonExisting', default=3.14)


class TestListField(resource_base.ListField):
    string = resource_base.Field('String', required=True)
    integer = resource_base.Field('Integer', adapter=int)


class TestDictionaryField(resource_base.DictionaryField):
    property_a = resource_base.Field('property_a')
    property_b = resource_base.Field('property_b')


class ComplexResource(resource_base.ResourceBase):
    string = resource_base.Field('String', required=True)
    integer = resource_base.Field('Integer', adapter=int)
    nested = NestedTestField('Nested')
    mapped_list = resource_base.MappedListField('MappedList', MAPPING)
    field_list = TestListField('ListField')
    dictionary = TestDictionaryField('Dictionary')
    non_existing_nested = NestedTestField('NonExistingNested')
    non_existing_mapped = resource_base.MappedField('NonExistingMapped',
                                                    MAPPING)
    enum_mapped = resource_base.MappedField('Enum', EnumMapping)
    enum_mapped_list = resource_base.MappedListField('EnumList', EnumMapping)
    non_existing_enum_mapped = resource_base.MappedField('NonExistingEnum',
                                                         EnumMapping)


class FieldTestCase(base.TestCase):
    def setUp(self):
        super(FieldTestCase, self).setUp()
        self.conn = mock.Mock()
        self.json = copy.deepcopy(TEST_JSON)
        self.conn.get.return_value.json.return_value = self.json
        self.test_resource = ComplexResource(self.conn,
                                             redfish_version='1.0.x')

    def test_ok(self):
        self.assertEqual('a string', self.test_resource.string)
        self.assertEqual(42, self.test_resource.integer)
        self.assertEqual('another string', self.test_resource.nested.string)
        self.assertEqual(0, self.test_resource.nested.integer)
        self.assertEqual('field value', self.test_resource.nested.nested_field)
        self.assertEqual('real', self.test_resource.nested.mapped)
        self.assertEqual(3.14, self.test_resource.nested.non_existing)
        self.assertEqual(['real1', 'real2', 'real'],
                         self.test_resource.mapped_list)
        self.assertEqual('a third string',
                         self.test_resource.field_list[0].string)
        self.assertEqual(2, self.test_resource.field_list[1].integer)
        self.assertEqual(2, len(self.test_resource.dictionary))
        self.assertEqual('value1',
                         self.test_resource.dictionary['key1'].property_a)
        self.assertEqual('value4',
                         self.test_resource.dictionary['key2'].property_b)
        self.assertIsNone(self.test_resource.non_existing_nested)
        self.assertIsNone(self.test_resource.non_existing_mapped)
        self.assertEqual(EnumMapping.FIELD2, self.test_resource.enum_mapped)
        self.assertEqual([EnumMapping.FIELD2, EnumMapping.FIELD3],
                         self.test_resource.enum_mapped_list)
        self.assertIsNone(self.test_resource.non_existing_enum_mapped)

    def test_missing_required(self):
        del self.json['String']
        self.assertRaisesRegex(
            exceptions.MissingAttributeError,
            'String', self.test_resource.refresh, force=True)

    def test_missing_nested_required(self):
        del self.json['Nested']['String']
        self.assertRaisesRegex(
            exceptions.MissingAttributeError,
            'Nested/String', self.test_resource.refresh, force=True)

    def test_missing_nested_required2(self):
        del self.json['Nested']['Object']['Field']
        self.assertRaisesRegex(exceptions.MissingAttributeError,
                               'Nested/Object/Field',
                               self.test_resource.refresh, force=True)

    def test_malformed_int(self):
        self.json['Integer'] = 'banana'
        self.assertRaisesRegex(
            exceptions.MalformedAttributeError,
            'attribute Integer is malformed.*invalid literal for int',
            self.test_resource.refresh, force=True)

    def test_malformed_nested_int(self):
        self.json['Nested']['Integer'] = 'banana'
        self.assertRaisesRegex(
            exceptions.MalformedAttributeError,
            'attribute Nested/Integer is malformed.*invalid literal for int',
            self.test_resource.refresh, force=True)

    def test_mapping_missing(self):
        self.json['Nested']['Mapped'] = 'banana'
        self.json['Enum'] = 'banana'
        self.test_resource.refresh(force=True)

        self.assertIsNone(self.test_resource.nested.mapped)
        self.assertIsNone(self.test_resource.enum_mapped)

    def test_composite_field_as_mapping(self):
        field = self.test_resource.nested
        keys = {'string', 'integer', 'nested_field', 'mapped', 'non_existing'}
        values = {'another string', 0, 'field value', 'real', 3.14}

        self.assertEqual(keys, set(iter(field)))
        self.assertEqual(keys, set(field.keys()))
        self.assertEqual(values, set(field.values()))
        self.assertEqual(3.14, field['non_existing'])
        self.assertEqual(3.14, field.get('non_existing'))
        self.assertIsNone(field.get('foobar'))
        # Check KeyError from undefined fields
        self.assertRaisesRegex(KeyError, 'foobar', lambda: field['foobar'])
        # Regular attributes cannot be accessed via mapping
        self.assertRaisesRegex(KeyError, '_load', lambda: field['_load'])
        self.assertRaisesRegex(KeyError, '__init__', lambda: field['__init__'])

    def test_invalid_mapping_definiton(self):
        self.assertRaises(TypeError, resource_base.MappedField, 'Field', 42)
        self.assertRaises(TypeError, resource_base.MappedListField,
                          'Field', 42)


class PartialKeyResource(resource_base.ResourceBase):
    string = resource_base.Field(
        lambda key, **context: key.startswith('Str'))
    integer = resource_base.Field(
        lambda key, value, **context: key == 'Integer' and int(value) < 42)


class FieldPartialKeyTestCase(base.TestCase):
    def setUp(self):
        super(FieldPartialKeyTestCase, self).setUp()
        self.conn = mock.Mock()
        self.json = copy.deepcopy(TEST_JSON)
        self.conn.get.return_value.json.return_value = self.json
        self.test_resource = PartialKeyResource(
            self.conn, redfish_version='1.0.x')

    def test_ok(self):
        self.assertEqual('a string', self.test_resource.string)
        self.assertIsNone(self.test_resource.integer)



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/Session.v1_1_0.json

import logging

from sushy.resources import base

LOG = logging.getLogger(__name__)


class Session(base.ResourceBase):

    description = base.Field('Description')
    """The session service description"""

    identity = base.Field('Id', required=True)
    """The session service identify string"""

    name = base.Field('Name', required=True)
    """The session service name"""

    username = base.Field('UserName')
    """The UserName for the account for this session."""

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None, root=None):
        """A class representing a Session

        :param connector: A Connector instance
        :param identity: The identity of the Session resource
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        super(Session, self).__init__(
            connector, identity, redfish_version=redfish_version,
            registries=registries, root=root)

    def delete(self):
        """Method for deleting a Session.

        :raises: ServerSideError
        """
        self._conn.delete(self.path)


class SessionCollection(base.ResourceCollectionBase):

    name = base.Field('Name')
    """The session collection name"""

    description = base.Field('Description')
    """The session collection description"""

    @property
    def _resource_type(self):
        return Session

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None, root=None):
        """A class representing a SessionCollection

        :param connector: A Connector instance
        :param identity: The identity of the Session resource
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        super(SessionCollection, self).__init__(
            connector, identity, redfish_version=redfish_version,
            registries=registries, root=root)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/v1/MessageRegistryFileCollection.json
# https://redfish.dmtf.org/schemas/v1/MessageRegistryFile.v1_1_0.json

import logging

from sushy.resources import base
from sushy.resources.registry import attribute_registry
from sushy.resources.registry import message_registry

LOG = logging.getLogger(__name__)


class LocationListField(base.ListField):
    """Location for each registry file of languages supported

    There are 3 options where the file can be hosted:

    * locally as a single file,
    * locally as a part of archive (zip or other),
    * publicly on the Internet.
    """

    language = base.Field('Language')
    """File's RFC5646 language code or the string 'default'"""

    uri = base.Field('Uri')
    """Location URI for co-located registry file with the Redfish service"""

    archive_uri = base.Field('ArchiveUri')
    """Location URI for  archive file"""

    archive_file = base.Field('ArchiveFile')
    """File name for registry if using archive_uri"""

    publication_uri = base.Field('PublicationUri')
    """Location URI of publicly available schema"""


class RegistryType(base.ResourceBase):
    _odata_type = base.Field('@odata.type', required=True)


class MessageRegistryFile(base.ResourceBase):

    identity = base.Field('Id', required=True)
    """Identity of Message Registry file resource"""

    description = base.Field('Description')
    """Description of Message Registry file resource"""

    name = base.Field('Name', required=True)
    """Name of Message Registry file resource"""

    languages = base.Field('Languages', required=True)
    """List of RFC 5646 language codes supported by this resource"""

    registry = base.Field('Registry', required=True, default='UNKNOWN.0.0')
    """Prefix for MessageId used for messages from this resource

    This attribute is in form Registry_name.Major_version.Minor_version
    """

    location = LocationListField('Location', required=True)
    """List of locations of Registry files for each supported language"""

    def get_message_registry(self, language, public_connector):
        """Get a Message Registry from the location

        :param language: RFC 5646 language code for registry files
        :param public_connector: connector to use when downloading registry
            from the Internet
        :returns: a MessageRegistry or None if not found
        """
        return self._get_registry(language, public_connector,
                                  'MessageRegistry',
                                  message_registry.MessageRegistry)

    def get_attribute_registry(self, language, public_connector):
        """Get an Attribute Registry from the location

        :param language: RFC 5646 language code for registry files
        :param public_connector: connector to use when downloading registry
            from the Internet
        :returns: an AttributeRegistry or None if not found
        """
        return self._get_registry(language, public_connector,
                                  'AttributeRegistry',
                                  attribute_registry.AttributeRegistry)

    def _get_registry(self, language, public_connector, requested_type,
                      registry_class):
        """Load registry file depending on the registry type

        Will try to find requested_type based on `odata.type` property,
        location, and provided language. If desired language is not found,
        will pick a registry that has 'default' language.

        :param language: RFC 5646 language code for registry files
        :param public_connector: connector to use when downloading registry
            from the Internet
        :param requested_type: string identifying registry
        :param registry_class: registry class
        :returns: registry or None if not found
        """

        # NOTE (etingof): as per RFC5646, languages are case-insensitive
        language = language.lower()

        locations = [
            l for l in self.location if l.language.lower() == language]

        locations += [
            l for l in self.location if l.language.lower() == 'default']

        for location in locations:
            if location.uri:
                args = self._conn,
                kwargs = {
                    'path': location.uri,
                    'reader': None,
                    'redfish_version': self.redfish_version
                }

            elif location.archive_uri:
                args = self._conn,
                kwargs = {
                    'path': location.archive_uri,
                    'reader': base.JsonArchiveReader(location.archive_file),
                    'redfish_version': self.redfish_version
                }

            elif location.publication_uri:
                args = public_connector,
                kwargs = {
                    'path': location.publication_uri,
                    'reader': base.JsonPublicFileReader(),
                    'redfish_version': self.redfish_version
                }

            else:
                LOG.warning('Incomplete location for language %(language)s',
                            {'language': language})
                continue

            try:
                registry_type = RegistryType(*args, **kwargs)

            except Exception as exc:
                LOG.warning(
                    'Cannot load registry type from location '
                    '%(location)s: %(error)s', {
                        'location': kwargs['path'],
                        'error': exc})
                continue

            if registry_type._odata_type.endswith(requested_type):
                try:
                    return registry_class(*args, **kwargs)

                except Exception as exc:
                    LOG.warning(
                        'Cannot load registry %(type)s from location '
                        '%(location)s: %(error)s', {
                            'type': requested_type,
                            'location': kwargs['path'],
                            'error': exc})
                    continue

            LOG.debug('Ignoring unsupported flavor of registry %(registry)s',
                      {'registry': registry_type._odata_type})
            return

        LOG.warning('No registry found for %(language)s or default',
                    {'language': language})


class MessageRegistryFileCollection(base.ResourceCollectionBase):
    """Collection of Message Registry Files"""

    @property
    def _resource_type(self):
        return MessageRegistryFile



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This is referred from Redfish standard schema.
# http://redfish.dmtf.org/schemas/v1/StorageController.v1_6_0.json

import logging

from sushy.resources import base
from sushy.resources import common
from sushy.resources import constants as res_cons
from sushy.resources import settings
from sushy.resources.system.storage import constants
from sushy.taskmonitor import TaskMonitor
from sushy import utils


LOG = logging.getLogger(__name__)


class StorageController(base.ResourceBase):
    """Storage controller"""

    identity = base.Field('Id', required=True)
    """The storage controller identity"""

    name = base.Field('Name', required=True)
    """The name of the storage controller"""

    status = common.StatusField('Status')
    """Describes the status and health of the resource and its children."""

    identifiers = common.IdentifiersListField('Identifiers', default=[])
    """The Durable names for the storage controller."""

    speed_gbps = base.Field('SpeedGbps')
    """The maximum speed of the storage controller's device interface."""

    controller_protocols = base.MappedListField(
        'SupportedControllerProtocols', res_cons.Protocol)
    """The protocols by which this storage controller can be communicated to"""

    device_protocols = base.MappedListField('SupportedDeviceProtocols',
                                            res_cons.Protocol)
    """The protocols that can be used to communicate with attached devices"""

    raid_types = base.MappedListField('SupportedRAIDTypes', constants.RAIDType)
    """The set of RAID types supported by the storage controller."""

    _settings = settings.SettingsField()
    """Future intended state for settings that can't be updated immediately."""

    @property
    @utils.cache_it
    def pending_settings(self):
        """Pending Storage Controller settings resource"""
        return StorageController(
            self._conn, self._settings.resource_uri,
            registries=None,
            redfish_version=self.redfish_version, root=self.root)

    @property
    def supported_apply_times(self):
        """List of supported BIOS update apply times

        :returns: List of supported update apply time names
        """
        return self._settings._supported_apply_times

    def update(self, payload, apply_time=None, maint_window_start_time=None,
               maint_window_duration=None):
        """Updates writable properties

        Supports updating properties that require reboot.

        :param payload: dictionary with properties to update
        :param apply_time: When to update the attributes. Optional.
            A :py:class:`sushy.ApplyTime` value.
        :param maint_window_start_time: The start time of a maintenance window,
            datetime. Required when updating during maintenance window and
            default maintenance window not set by the system.
        :param maint_window_duration: Duration of maintenance time since
            maintenance window start time in seconds. Required when updating
            during maintenance window and default maintenance window not
            set by the system.
        :returns: TaskMonitor if async task or None
        """
        payload = utils.process_apply_time_input(
            payload, apply_time, maint_window_start_time,
            maint_window_duration)
        r = self._settings.commit(self._conn, payload)
        utils.cache_clear(self, force_refresh=False,
                          only_these=['pending_settings'])
        if r.status_code == 202:
            return TaskMonitor.from_response(
                self._conn, r, self._settings.resource_uri,
                self.redfish_version, self.registries)


class ControllerCollection(base.ResourceCollectionBase):

    @property
    def _resource_type(self):
        return StorageController

    @property
    @utils.cache_it
    def summary(self):
        """Summary of storage controllers

        :returns: dictionary of controller id-s and their status in format

        .. code-block:: python

          {'RAID.Integrated.1-1': {'Health': sushy.Health.OK,
                                   'State': sushy.State.ENABLED}}
        """
        controllers = {}
        for controller in self.get_members():
            controllers[controller.identity] = {
                'Health': controller.status.health,
                'State': controller.status.state,
            }
        return controllers

    def __init__(self, connector, path, redfish_version=None, registries=None,
                 root=None):
        """A class representing a ControllerCollection

        :param connector: A Connector instance
        :param path: The canonical path to the Controller collection resource
        :param redfish_version: The version of Redfish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        super(ControllerCollection, self).__init__(
            connector, path, redfish_version=redfish_version,
            registries=registries,
            root=root)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Values come from the Redfish json-schema:
# https://redfish.dmtf.org/schemas/v1/IPAddresses.v1_1_3.json

import enum


class AddressState(enum.Enum):
    PREFERRED = 'Preferred'
    """This address is currently within both its RFC4862-defined valid and
    preferred lifetimes."""

    DEPRECATED = 'Deprecated'
    """This address is currently within its valid lifetime but is now
    outside its RFC4862-defined preferred lifetime."""

    TENTATIVE = 'Tentative'
    """This address is currently undergoing Duplicate Address Detection
    (DAD) testing, as defined in RFC4862, section 5.4."""

    FAILED = 'Failed'
    """This address has failed Duplicate Address Detection (DAD) testing, as
    defined in RFC4862, section 5.4, and is not currently in use."""


# Backward compatibility
ADDRESS_STATE_PREFERRED = AddressState.PREFERRED
ADDRESS_STATE_DEPRECATED = AddressState.DEPRECATED
ADDRESS_STATE_TENTATIVE = AddressState.TENTATIVE
ADDRESS_STATE_FAILED = AddressState.FAILED


class IPv4AddressOrigin(enum.Enum):
    STATIC = 'Static'
    """A user-configured static address."""

    DHCP = 'DHCP'
    """A DHCPv4 service-provided address."""

    BOOTP = 'BOOTP'
    """A BOOTP service-provided address."""

    LINK_LOCAL = 'IPv4LinkLocal'
    """The address is valid for only this network segment, or link."""


# Backward compatibility
ADDRESS_ORIGIN_IPv4_BOOTP = IPv4AddressOrigin.BOOTP
ADDRESS_ORIGIN_IPv4_DHCP = IPv4AddressOrigin.DHCP
ADDRESS_ORIGIN_IPv4_IPv4LINKLOCAL = IPv4AddressOrigin.LINK_LOCAL
ADDRESS_ORIGIN_IPv4_STATIC = IPv4AddressOrigin.STATIC


class IPv6AddressOrigin(enum.Enum):
    STATIC = 'Static'
    """A static user-configured address."""

    DHCP = 'DHCPv6'
    """A DHCPv6 service-provided address."""

    LINK_LOCAL = 'LinkLocal'
    """The address is valid for only this network segment, or link."""

    SLAAC = 'SLAAC'
    """A stateless autoconfiguration (SLAAC) service-provided address."""


# Backward compatibility
ADDRESS_ORIGIN_IPv6_DHCPv6 = IPv6AddressOrigin.DHCP
ADDRESS_ORIGIN_IPv6_LINKLOCAL = IPv6AddressOrigin.LINK_LOCAL
ADDRESS_ORIGIN_IPv6_SLAAC = IPv6AddressOrigin.SLAAC
ADDRESS_ORIGIN_IPv6_STATIC = IPv6AddressOrigin.STATIC



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/v1/TaskService.v1_1_5.json

import logging

from sushy.resources import base
from sushy.resources import common
from sushy.resources.taskservice import constants as ts_cons
from sushy.resources.taskservice import task
from sushy import utils

LOG = logging.getLogger(__name__)


class TaskService(base.ResourceBase):

    identity = base.Field('Id', required=True)
    """The task service identity"""

    name = base.Field('Name', required=True)
    """The task service name"""

    service_enabled = base.Field('ServiceEnabled')
    """The status of whether this service is enabled"""

    status = common.StatusField('Status')
    """The status of the task service"""

    overwrite_policy = base.MappedField(
        'CompletedTaskOverWritePolicy', ts_cons.OverWritePolicy)
    """The overwrite policy for completed tasks"""

    event_on_task_state_change = base.Field(
        'LifeCycleEventOnTaskStateChange', adapter=bool)
    """Whether a task state change sends an event"""

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None, root=None):
        """A class representing a TaskService

        :param connector: A Connector instance
        :param identity: The identity of the TaskService resource
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of given version
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        super(TaskService, self).__init__(
            connector, identity, redfish_version=redfish_version,
            registries=registries, root=root)

    @property
    @utils.cache_it
    def tasks(self):
        """Property to reference `TaskCollection` instance

        It is set once when the first time it is queried. On refresh,
        this property is marked as stale (greedy-refresh not done).
        Here the actual refresh of the sub-resource happens, if stale.
        """
        return task.TaskCollection(
            self._conn, utils.get_sub_resource_path_by(self, 'Tasks'),
            redfish_version=self.redfish_version,
            registries=self.registries, root=self.root)



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/v1/ComputerSystem.v1_10_0.json

import collections
import logging

from sushy import exceptions
from sushy.resources import base
from sushy.resources.chassis import chassis
from sushy.resources import common
from sushy.resources import constants as res_cons
from sushy.resources.manager import manager
from sushy.resources import settings
from sushy.resources.system import bios
from sushy.resources.system import constants as sys_cons
from sushy.resources.system import ethernet_interface
from sushy.resources.system import processor
from sushy.resources.system import secure_boot
from sushy.resources.system import simple_storage as sys_simple_storage
from sushy.resources.system.storage import storage as sys_storage
from sushy import utils


LOG = logging.getLogger(__name__)


class ActionsField(base.CompositeField):
    reset = common.ResetActionField('#ComputerSystem.Reset')


class BootField(base.CompositeField):
    allowed_values = base.Field(
        'BootSourceOverrideTarget@Redfish.AllowableValues',
        adapter=list)

    enabled = base.MappedField('BootSourceOverrideEnabled',
                               sys_cons.BootSourceOverrideEnabled)

    mode = base.MappedField('BootSourceOverrideMode',
                            sys_cons.BootSourceOverrideMode)

    target = base.MappedField('BootSourceOverrideTarget', sys_cons.BootSource)


class MemorySummaryField(base.CompositeField):
    health = base.Field(['Status', 'HealthRollup'])
    """The overall health state of memory.

    This signifies health state of memory along with its dependent resources.
    """

    size_gib = base.Field('TotalSystemMemoryGiB', adapter=utils.int_or_none)
    """The size of memory of the system in GiB.

    This signifies the total installed, operating system-accessible memory
    (RAM), measured in GiB.
    """


class System(base.ResourceBase):

    asset_tag = base.Field('AssetTag')
    """The system asset tag"""

    bios_version = base.Field('BiosVersion')
    """The system BIOS version"""

    boot = BootField('Boot', required=True)
    """A dictionary containing the current boot device, frequency and mode"""

    description = base.Field('Description')
    """The system description"""

    hostname = base.Field('HostName')
    """The system hostname"""

    identity = base.Field('Id', required=True)
    """The system identity string"""

    indicator_led = base.MappedField('IndicatorLED', res_cons.IndicatorLED)
    """Whether the indicator LED is lit or off"""

    manufacturer = base.Field('Manufacturer')
    """The system manufacturer"""

    name = base.Field('Name')
    """The system name"""

    part_number = base.Field('PartNumber')
    """The system part number"""

    power_state = base.MappedField('PowerState', res_cons.PowerState)
    """The system power state"""

    serial_number = base.Field('SerialNumber')
    """The system serial number"""

    sku = base.Field('SKU')
    """The system stock-keeping unit"""

    status = common.StatusField('Status')
    """The system status"""

    system_type = base.MappedField('SystemType', sys_cons.SystemType)
    """The system type"""

    uuid = base.Field('UUID')
    """The system UUID"""

    memory_summary = MemorySummaryField('MemorySummary')
    """The summary info of memory of the system in general detail"""

    maintenance_window = settings.MaintenanceWindowField(
        '@Redfish.MaintenanceWindow')
    """Indicates if a given resource has a maintenance window assignment
    for applying settings or operations"""

    _settings = settings.SettingsField()
    """Settings Resource is used to represent the future intended state
    of a Resource
    Ref: http://redfish.dmtf.org/schemas/DSP0266_1.7.0.html#settings-resource
    """

    _actions = ActionsField('Actions', required=True)

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None, root=None):
        """A class representing a ComputerSystem

        :param connector: A Connector instance
        :param identity: The identity of the System resource
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of registries to be used in any resource
            that needs registries to parse messages.
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        super(System, self).__init__(
            connector, identity,
            redfish_version=redfish_version,
            registries=registries,
            root=root)

    def _get_reset_action_element(self):
        reset_action = self._actions.reset
        # TODO(dtantsur): make this check also declarative?
        if not reset_action:
            raise exceptions.MissingActionError(action='#ComputerSystem.Reset',
                                                resource=self._path)
        return reset_action

    def get_allowed_reset_system_values(self):
        """Get the allowed values for resetting the system.

        :returns: A set with the allowed values.
        """
        reset_action = self._get_reset_action_element()

        if not reset_action.allowed_values:
            LOG.warning('Could not figure out the allowed values for the '
                        'reset system action for System %s', self.identity)
            return set(res_cons.ResetType)

        return {v for v in res_cons.ResetType
                if v.value in reset_action.allowed_values}

    def reset_system(self, value):
        """Reset the system.

        :param value: The target value.
        :raises: InvalidParameterValueError, if the target value is not
            allowed.
        """
        valid_resets = self.get_allowed_reset_system_values()
        if value not in valid_resets:
            raise exceptions.InvalidParameterValueError(
                parameter='value', value=value, valid_values=valid_resets)

        value = res_cons.ResetType(value).value
        target_uri = self._get_reset_action_element().target_uri

        # TODO(lucasagomes): Check the return code and response body ?
        #                    Probably we should call refresh() as well.
        self._conn.post(target_uri, data={'ResetType': value})

    def get_allowed_system_boot_source_values(self):
        """Get the allowed values for changing the boot source.

        :returns: A set with the allowed values.
        """
        if not self.boot.allowed_values:
            LOG.warning('Could not figure out the allowed values for '
                        'configuring the boot source for System %s',
                        self.identity)
            return set(sys_cons.BootSource)

        return {v for v in sys_cons.BootSource
                if v.value in self.boot.allowed_values}

    def set_system_boot_options(self, target=None, enabled=None, mode=None):
        """Set boot source and/or boot frequency and/or boot mode.

        Set the boot source and/or boot frequency and/or boot mode to use
        on next reboot of the System.

        :param target: The target boot source,
            a :py:class:`sushy.BootSource` value. Optional.
        :param enabled: How long the override be enabled,
            a :py:class:`sushy.BootSourceOverrideEnabled` value. Optional.
        :param mode: The boot mode,
            a :py:class:`sushy.BootSourceOverrideMode` value. Optional.
        :raises: InvalidParameterValueError, if any information passed is
            invalid.
        """
        data = collections.defaultdict(dict)
        settings_data = collections.defaultdict(dict)

        if self._settings and self._settings.resource_uri:
            settings_resp = self._conn.get(self._settings.resource_uri)
            settings_boot_section = settings_resp.json().get('Boot', {})
        else:
            settings_resp = None

        if target is not None:
            valid_targets = self.get_allowed_system_boot_source_values()
            if target not in valid_targets:
                raise exceptions.InvalidParameterValueError(
                    parameter='target', value=target,
                    valid_values=valid_targets)

            target = sys_cons.BootSource(target)
            # NOTE(janders) on SuperMicro X11 and X12 machines, virtual media
            # is presented as an "USB CD" drive as opposed to a CD drive. Both
            # are present in the list of boot devices, however only selecting
            # UsbCd as the boot source results in a successful boot from
            # vMedia. If "CD" is selected, boot fails even if vMedia is
            # inserted. This code detects a case where a SuperMicro machine is
            # about to attempt boot from CD and overrides the boot device to
            # UsbCd instead which makes boot from vMedia work as expected.
            if (self.manufacturer and self.manufacturer.lower() == 'supermicro'
                    and target == sys_cons.BootSource.CD
                    and sys_cons.BootSource.USB_CD.value
                    in self.boot.allowed_values):
                LOG.debug('Boot from vMedia was requested on a SuperMicro'
                          'machine. Overriding boot device from %s to %s.',
                          target, sys_cons.BootSource.USB_CD)
                target = sys_cons.BootSource.USB_CD
            if (settings_resp and "BootSourceOverrideTarget" in
                    settings_boot_section):
                settings_data['Boot']['BootSourceOverrideTarget'] = \
                    target.value
            else:
                data['Boot']['BootSourceOverrideTarget'] = target.value

        if enabled is not None:
            try:
                fishy_freq = sys_cons.BootSourceOverrideEnabled(enabled).value
            except ValueError:
                raise exceptions.InvalidParameterValueError(
                    parameter='enabled', value=enabled,
                    valid_values=list(sys_cons.BootSourceOverrideEnabled))
            if (settings_resp and "BootSourceOverrideEnabled" in
                    settings_boot_section):
                settings_data['Boot']['BootSourceOverrideEnabled'] = fishy_freq
            else:
                data['Boot']['BootSourceOverrideEnabled'] = fishy_freq

        if mode is not None:
            try:
                fishy_mode = sys_cons.BootSourceOverrideMode(mode).value
            except ValueError:
                raise exceptions.InvalidParameterValueError(
                    parameter='mode', value=mode,
                    valid_values=list(sys_cons.BootSourceOverrideMode))
            if (settings_resp and "BootSourceOverrideMode" in
                    settings_boot_section):
                settings_data['Boot']['BootSourceOverrideMode'] = fishy_mode
            else:
                data['Boot']['BootSourceOverrideMode'] = fishy_mode

        # TODO(lucasagomes): Check the return code and response body ?
        #                    Probably we should call refresh() as well.
        if settings_data.get('Boot'):
            etag = settings_resp.headers.get('ETag')
            path = self._settings.resource_uri
            self._conn.patch(path, data=settings_data,
                             etag=etag)
        if data.get('Boot'):
            etag = self._get_etag()
            path = self.path
            self._conn.patch(path, data=data, etag=etag)

    # TODO(etingof): we should remove this method, eventually
    def set_system_boot_source(
            self, target, enabled=sys_cons.BootSourceOverrideEnabled.ONCE,
            mode=None):
        """Set boot source and/or boot frequency and/or boot mode.

        Set the boot source and/or boot frequency and/or boot mode to use
        on next reboot of the System.

        This method is obsoleted by `set_system_boot_options`.

        :param target: The target boot source,
            a :py:class:`sushy.BootSource` value.
        :param enabled: The frequency, whether to set it for the next
            a :py:class:`sushy.BootSourceOverrideEnabled` value.
            Default is `ONCE`.
        :param mode: The boot mode,
            a :py:class:`sushy.BootSourceOverrideMode` value.
        :raises: InvalidParameterValueError, if any information passed is
            invalid.
        """
        self.set_system_boot_options(target, enabled, mode)

    def set_indicator_led(self, state):
        """Set IndicatorLED to the given state.

        :param state: Desired LED state, an IndicatorLED value.
        :raises: InvalidParameterValueError, if any information passed is
            invalid.
        """
        try:
            state = res_cons.IndicatorLED(state).value
        except ValueError:
            raise exceptions.InvalidParameterValueError(
                parameter='state', value=state,
                valid_values=' ,'.join(i.value for i in res_cons.IndicatorLED))

        data = {'IndicatorLED': state}

        self._conn.patch(self.path, data=data)
        self.invalidate()

    def _get_processor_collection_path(self):
        """Helper function to find the ProcessorCollection path"""
        return utils.get_sub_resource_path_by(self, 'Processors')

    @property
    @utils.cache_it
    def processors(self):
        """Property to reference `ProcessorCollection` instance

        It is set once when the first time it is queried. On refresh,
        this property is marked as stale (greedy-refresh not done).
        Here the actual refresh of the sub-resource happens, if stale.
        """
        return processor.ProcessorCollection(
            self._conn, self._get_processor_collection_path(),
            redfish_version=self.redfish_version,
            registries=self.registries, root=self.root)

    @property
    @utils.cache_it
    def ethernet_interfaces(self):
        """Property to reference `EthernetInterfaceCollection` instance

        It is set once when the first time it is queried. On refresh,
        this property is marked as stale (greedy-refresh not done).
        Here the actual refresh of the sub-resource happens, if stale.
        """
        return ethernet_interface.EthernetInterfaceCollection(
            self._conn,
            utils.get_sub_resource_path_by(self, "EthernetInterfaces"),
            redfish_version=self.redfish_version,
            registries=self.registries, root=self.root)

    @property
    @utils.cache_it
    def bios(self):
        """Property to reference `Bios` instance

        It is set once when the first time it is queried. On refresh,
        this property is marked as stale (greedy-refresh not done).
        Here the actual refresh of the sub-resource happens, if stale.
        """
        return bios.Bios(
            self._conn,
            utils.get_sub_resource_path_by(self, 'Bios'),
            redfish_version=self.redfish_version,
            registries=self.registries, root=self.root)

    @property
    @utils.cache_it
    def simple_storage(self):
        """A collection of simple storage associated with system.

        This returns a reference to `SimpleStorageCollection` instance.
        SimpleStorage represents the properties of a storage controller and its
        directly-attached devices.

        It is set once when the first time it is queried. On refresh,
        this property is marked as stale (greedy-refresh not done).
        Here the actual refresh of the sub-resource happens, if stale.

        :raises: MissingAttributeError if 'SimpleStorage/@odata.id' field
            is missing.
        :returns: `SimpleStorageCollection` instance
        """
        return sys_simple_storage.SimpleStorageCollection(
            self._conn, utils.get_sub_resource_path_by(self, "SimpleStorage"),
            redfish_version=self.redfish_version,
            registries=self.registries, root=self.root)

    @property
    @utils.cache_it
    def storage(self):
        """A collection of storage subsystems associated with system.

        This returns a reference to `StorageCollection` instance.
        A storage subsystem represents a set of storage controllers (physical
        or virtual) and the resources such as drives and volumes that can be
        accessed from that subsystem.

        It is set once when the first time it is queried. On refresh,
        this property is marked as stale (greedy-refresh not done).
        Here the actual refresh of the sub-resource happens, if stale.

        :raises: MissingAttributeError if 'Storage/@odata.id' field
            is missing.
        :returns: `StorageCollection` instance
        """
        return sys_storage.StorageCollection(
            self._conn, utils.get_sub_resource_path_by(self, "Storage"),
            redfish_version=self.redfish_version,
            registries=self.registries, root=self.root)

    @property
    @utils.cache_it
    def managers(self):
        """A list of managers for this system.

        Returns a list of `Manager` objects representing the managers
        that manage this system.

        :raises: MissingAttributeError if '@odata.id' field is missing.
        :returns: A list of `Manager` instances
        """
        paths = utils.get_sub_resource_path_by(
            self, ["Links", "ManagedBy"], is_collection=True)

        return [manager.Manager(self._conn, path,
                                redfish_version=self.redfish_version,
                                registries=self.registries, root=self.root)
                for path in paths]

    @property
    @utils.cache_it
    def chassis(self):
        """A list of chassis where this system resides.

        Returns a list of `Chassis` objects representing the chassis
        or cabinets where this system is mounted.

        :raises: MissingAttributeError if '@odata.id' field is missing.
        :returns: A list of `Chassis` instances
        """
        paths = utils.get_sub_resource_path_by(
            self, ["Links", "Chassis"], is_collection=True)

        return [chassis.Chassis(self._conn, path,
                                redfish_version=self.redfish_version,
                                registries=self.registries, root=self.root)
                for path in paths]

    @property
    @utils.cache_it
    def secure_boot(self):
        """Property to reference `SecureBoot` instance

        It is set once when the first time it is queried. On refresh,
        this property is marked as stale (greedy-refresh not done).
        Here the actual refresh of the sub-resource happens, if stale.
        """
        return secure_boot.SecureBoot(
            self._conn,
            utils.get_sub_resource_path_by(self, 'SecureBoot'),
            redfish_version=self.redfish_version,
            registries=self.registries, root=self.root)


class SystemCollection(base.ResourceCollectionBase):

    @property
    def _resource_type(self):
        return System

    def __init__(self, connector, path, redfish_version=None, registries=None,
                 root=None):
        """A class representing a ComputerSystemCollection

        :param connector: A Connector instance
        :param path: The canonical path to the System collection resource
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        super(SystemCollection, self).__init__(
            connector, path, redfish_version=redfish_version,
            registries=registries, root=root)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client
import json
from unittest import mock

import sushy
from sushy.resources.system.storage import controller
from sushy import taskmonitor
from sushy.tests.unit import base


class ControllerTestCase(base.TestCase):

    def setUp(self):
        super(ControllerTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'storage_controller.json') as f:
            self.json_doc = json.load(f)
        with open('sushy/tests/unit/json_samples/'
                  'storage_controller_settings.json') as f:
            self.settings_json = json.load(f)

        self.conn.get.return_value.json.side_effect = [
            self.json_doc, self.settings_json]

        self.controller = controller.StorageController(
            self.conn,
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Controllers/1',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.controller._parse_attributes(self.json_doc)
        self.assertEqual('1', self.controller.identity)
        self.assertEqual('Contoso Integrated RAID', self.controller.name)
        self.assertEqual(sushy.Health.OK, self.controller.status.health)
        self.assertEqual(sushy.State.ENABLED, self.controller.status.state)
        identifiers = self.controller.identifiers
        self.assertIsInstance(identifiers, list)
        self.assertEqual(1, len(identifiers))
        identifier = identifiers[0]
        self.assertEqual(sushy.DurableNameFormat.NAA,
                         identifier.durable_name_format)
        self.assertEqual('345C59DBD970859C', identifier.durable_name)
        self.assertEqual(12, self.controller.speed_gbps)
        self.assertEqual([sushy.Protocol.PCIe],
                         self.controller.controller_protocols)
        self.assertEqual([sushy.Protocol.SAS, sushy.Protocol.SATA],
                         self.controller.device_protocols)
        self.assertEqual([sushy.RAIDType.RAID0, sushy.RAIDType.RAID1],
                         self.controller.raid_types)
        self.assertEqual('1', self.controller.pending_settings.identity)
        self.assertEqual([sushy.ApplyTime.IMMEDIATE, sushy.ApplyTime.ON_RESET],
                         self.controller.supported_apply_times)

    def test_update(self):
        mock_response = mock.Mock()
        mock_response.status_code = http_client.ACCEPTED
        mock_response.headers = {'Content-Length': 42,
                                 'Location': '/Task/545',
                                 'Retry-After': 20,
                                 'Allow': 'DELETE'}
        mock_response.content = None
        self.controller._conn.patch.return_value = mock_response
        tm = self.controller.update(
            {'ControllerRates': {'ConsistencyCheckRatePercent': 30}},
            apply_time=sushy.ApplyTime.ON_RESET)
        self.controller._conn.patch.assert_called_once_with(
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Controllers/1/Settings',
            data={'ControllerRates': {'ConsistencyCheckRatePercent': 30},
                  '@Redfish.SettingsApplyTime': {
                      '@odata.type': '#Settings.v1_0_0.PreferredApplyTime',
                      'ApplyTime': 'OnReset'}})
        self.assertIsInstance(tm, taskmonitor.TaskMonitor)
        self.assertEqual('/Task/545', tm.task_monitor_uri)


class ControllerCollectionTestCase(base.TestCase):

    def setUp(self):
        super(ControllerCollectionTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'storage_controller_collection.json') as f:
            self.json_doc = json.load(f)

        with open('sushy/tests/unit/json_samples/'
                  'storage_controller.json') as f:
            self.json_doc_ctrl = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.ctrl_col = controller.ControllerCollection(
            self.conn,
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Controllers',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.ctrl_col._parse_attributes(self.json_doc)
        self.assertEqual((
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Controllers/1',),
            self.ctrl_col.members_identities)

    @mock.patch.object(controller, 'StorageController', autospec=True)
    def test_get_member(self, mock_controller):
        self.ctrl_col.get_member(
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Controllers/1')
        mock_controller.assert_called_once_with(
            self.ctrl_col._conn,
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Controllers/1',
            redfish_version=self.ctrl_col.redfish_version, registries=None,
            root=self.ctrl_col.root)

    @mock.patch.object(controller, 'StorageController', autospec=True)
    def test_get_members(self, mock_controller):
        members = self.ctrl_col.get_members()
        mock_controller.assert_called_once_with(
            self.ctrl_col._conn,
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Controllers/1',
            redfish_version=self.ctrl_col.redfish_version, registries=None,
            root=self.ctrl_col.root)
        self.assertIsInstance(members, list)
        self.assertEqual(1, len(members))

    def test_summary(self):
        self.conn.get.return_value.json.return_value = self.json_doc_ctrl
        self.assertEqual({'1': {'Health': sushy.Health.OK,
                                'State': sushy.State.ENABLED}},
                         self.ctrl_col.summary)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock


import sushy
from sushy import exceptions
from sushy.resources.chassis import chassis
from sushy.resources.manager import manager
from sushy.resources.manager import virtual_media
from sushy.resources.system import system
from sushy.tests.unit import base


class ManagerTestCase(base.TestCase):

    def setUp(self):
        super(ManagerTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/manager.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.manager = manager.Manager(self.conn, '/redfish/v1/Managers/BMC',
                                       redfish_version='1.0.2')

    def test__parse_attributes(self):
        # | WHEN |
        self.manager._parse_attributes(self.json_doc)
        # | THEN |
        self.assertEqual('1.0.2', self.manager.redfish_version)
        self.assertEqual('1.00', self.manager.firmware_version)
        self.assertFalse(self.manager.auto_dst_enabled)
        self.assertEqual(True, self.manager.graphical_console.service_enabled)
        self.assertEqual(
            2, self.manager.graphical_console.max_concurrent_sessions)
        self.assertEqual(True, self.manager.serial_console.service_enabled)
        self.assertEqual(
            1, self.manager.serial_console.max_concurrent_sessions)
        self.assertEqual(True, self.manager.command_shell.service_enabled)
        self.assertEqual(
            4, self.manager.command_shell.max_concurrent_sessions)
        self.assertEqual('Contoso BMC', self.manager.description)
        self.assertEqual('BMC', self.manager.identity)
        self.assertEqual('Manager', self.manager.name)
        self.assertEqual('Joo Janta 200', self.manager.model)
        self.assertEqual(sushy.ManagerType.BMC, self.manager.manager_type)
        self.assertEqual('58893887-8974-2487-2389-841168418919',
                         self.manager.uuid)

    def test_get_supported_graphical_console_types(self):
        # | GIVEN |
        expected = set([sushy.GraphicalConnectType.KVMIP])
        # | WHEN |
        values = self.manager.get_supported_graphical_console_types()
        # | THEN |
        self.assertEqual(expected, values)
        self.assertIsInstance(values, set)

    def test_get_supported_graphical_console_types_for_no_connect_types(self):
        # | GIVEN |
        graphical_console = self.manager.graphical_console
        expected = set([sushy.GraphicalConnectType.KVMIP,
                        sushy.GraphicalConnectType.OEM])

        for val in [None, []]:
            graphical_console.connect_types_supported = val
            # | WHEN |
            values = self.manager.get_supported_graphical_console_types()
            # | THEN |
            self.assertEqual(expected, values)
            self.assertIsInstance(values, set)

    def test_get_supported_graphical_console_types_missing_graphcon_attr(self):
        # | GIVEN |
        self.manager.graphical_console = None
        expected = set([sushy.GraphicalConnectType.KVMIP,
                        sushy.GraphicalConnectType.OEM])
        # | WHEN |
        values = self.manager.get_supported_graphical_console_types()
        # | THEN |
        self.assertEqual(expected, values)
        self.assertIsInstance(values, set)

    def test_get_supported_serial_console_types(self):
        # | GIVEN |
        expected = set([sushy.SerialConnectType.SSH,
                        sushy.SerialConnectType.TELNET,
                        sushy.SerialConnectType.IPMI])
        # | WHEN |
        values = self.manager.get_supported_serial_console_types()
        # | THEN |
        self.assertEqual(expected, values)
        self.assertIsInstance(values, set)

    def test_get_supported_serial_console_types_for_no_connect_types(self):
        # | GIVEN |
        serial_console = self.manager.serial_console
        expected = set([sushy.SerialConnectType.SSH,
                        sushy.SerialConnectType.TELNET,
                        sushy.SerialConnectType.IPMI,
                        sushy.SerialConnectType.OEM])

        for val in [None, []]:
            serial_console.connect_types_supported = val
            # | WHEN |
            values = self.manager.get_supported_serial_console_types()
            # | THEN |
            self.assertEqual(expected, values)
            self.assertIsInstance(values, set)

    def test_get_supported_serial_console_types_missing_serialcon_attr(self):
        # | GIVEN |
        self.manager.serial_console = None
        expected = set([sushy.SerialConnectType.SSH,
                        sushy.SerialConnectType.TELNET,
                        sushy.SerialConnectType.IPMI,
                        sushy.SerialConnectType.OEM])
        # | WHEN |
        values = self.manager.get_supported_serial_console_types()
        # | THEN |
        self.assertEqual(expected, values)
        self.assertIsInstance(values, set)

    def test_get_supported_command_shell_types(self):
        # | GIVEN |
        expected = set([sushy.CommandConnectType.SSH,
                        sushy.CommandConnectType.TELNET])
        # | WHEN |
        values = self.manager.get_supported_command_shell_types()
        # | THEN |
        self.assertEqual(expected, values)
        self.assertIsInstance(values, set)

    def test_get_supported_command_shell_types_for_no_connect_types(self):
        # | GIVEN |
        command_shell = self.manager.command_shell
        expected = set([sushy.CommandConnectType.SSH,
                        sushy.CommandConnectType.TELNET,
                        sushy.CommandConnectType.IPMI,
                        sushy.CommandConnectType.OEM])

        for val in [None, []]:
            command_shell.connect_types_supported = val
            # | WHEN |
            values = self.manager.get_supported_command_shell_types()
            # | THEN |
            self.assertEqual(expected, values)
            self.assertIsInstance(values, set)

    def test_get_supported_command_shell_types_missing_cmdshell_attr(self):
        # | GIVEN |
        self.manager.command_shell = None
        expected = set([sushy.CommandConnectType.SSH,
                        sushy.CommandConnectType.TELNET,
                        sushy.CommandConnectType.IPMI,
                        sushy.CommandConnectType.OEM])
        # | WHEN |
        values = self.manager.get_supported_command_shell_types()
        # | THEN |
        self.assertEqual(expected, values)
        self.assertIsInstance(values, set)

    def test_get_allowed_reset_manager_values(self):
        # | GIVEN |
        expected = set([sushy.ResetType.GRACEFUL_RESTART,
                        sushy.ResetType.FORCE_RESTART])
        # | WHEN |
        values = self.manager.get_allowed_reset_manager_values()
        # | THEN |
        self.assertEqual(expected, values)
        self.assertIsInstance(values, set)

    def test_get_allowed_reset_manager_values_for_no_values_set(self):
        # | GIVEN |
        self.manager._actions.reset.allowed_values = []
        expected = set([sushy.ResetType.GRACEFUL_SHUTDOWN,
                        sushy.ResetType.GRACEFUL_RESTART,
                        sushy.ResetType.FORCE_RESTART,
                        sushy.ResetType.FORCE_OFF,
                        sushy.ResetType.FORCE_ON,
                        sushy.ResetType.ON,
                        sushy.ResetType.NMI,
                        sushy.ResetType.PUSH_POWER_BUTTON,
                        sushy.ResetType.POWER_CYCLE,
                        sushy.ResetType.SUSPEND,
                        sushy.ResetType.RESUME,
                        sushy.ResetType.PAUSE])
        # | WHEN |
        values = self.manager.get_allowed_reset_manager_values()
        # | THEN |
        self.assertEqual(expected, values)
        self.assertIsInstance(values, set)

    def test_get_allowed_reset_manager_values_missing_action_reset_attr(self):
        # | GIVEN |
        self.manager._actions.reset = None
        # | WHEN & THEN |
        self.assertRaisesRegex(
            exceptions.MissingActionError, 'action #Manager.Reset',
            self.manager.get_allowed_reset_manager_values)

    def test_reset_manager(self):
        self.manager.reset_manager(sushy.ResetType.GRACEFUL_RESTART)
        self.manager._conn.post.assert_called_once_with(
            '/redfish/v1/Managers/BMC/Actions/Manager.Reset',
            data={'ResetType': 'GracefulRestart'})

    def test_reset_manager_with_invalid_value(self):
        self.assertRaises(exceptions.InvalidParameterValueError,
                          self.manager.reset_manager, 'invalid-value')

    def test_virtual_media(self):
        # | GIVEN |
        with open('sushy/tests/unit/json_samples/'
                  'virtual_media_collection.json') as f:
            virtual_media_collection_return_value = json.load(f)

        with open('sushy/tests/unit/json_samples/'
                  'virtual_media.json') as f:
            virtual_media_return_value = json.load(f)

        self.conn.get.return_value.json.side_effect = [
            virtual_media_collection_return_value, virtual_media_return_value]

        # | WHEN |
        actual_virtual_media = self.manager.virtual_media

        # | THEN |
        self.assertIsInstance(actual_virtual_media,
                              virtual_media.VirtualMediaCollection)
        self.assertEqual(actual_virtual_media.name, 'Virtual Media Services')

        member = actual_virtual_media.get_member(
            '/redfish/v1/Managers/BMC/VirtualMedia/Floppy1')

        self.assertEqual(member.image_name, "Sardine2.1.43.35.6a")
        self.assertTrue(member.inserted)
        self.assertFalse(member.write_protected)

    def test_virtual_media_on_refresh(self):
        # | GIVEN |
        with open('sushy/tests/unit/json_samples/'
                  'virtual_media_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        # | WHEN & THEN |
        vrt_media = self.manager.virtual_media
        self.assertIsInstance(vrt_media, virtual_media.VirtualMediaCollection)

        # On refreshing the manager instance...
        with open('sushy/tests/unit/json_samples/manager.json', 'r') as f:
            self.conn.get.return_value.json.return_value = json.loads(f.read())

        self.manager.invalidate()
        self.manager.refresh(force=False)

        # | WHEN & THEN |
        self.assertTrue(vrt_media._is_stale)

        # | GIVEN |
        with open('sushy/tests/unit/json_samples/'
                  'virtual_media_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        # | WHEN & THEN |
        self.assertIsInstance(self.manager.virtual_media,
                              virtual_media.VirtualMediaCollection)
        self.assertFalse(vrt_media._is_stale)

    def test_systems(self):
        # | GIVEN |
        with open('sushy/tests/unit/json_samples/'
                  'system.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        # | WHEN & THEN |
        actual_systems = self.manager.systems
        self.assertIsInstance(actual_systems[0], system.System)
        self.assertEqual(
            '/redfish/v1/Systems/437XR1138R2', actual_systems[0].path)

    def test_chassis(self):
        # | GIVEN |
        with open('sushy/tests/unit/json_samples/'
                  'chassis.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        # | WHEN & THEN |
        actual_chassis = self.manager.chassis
        self.assertIsInstance(actual_chassis[0], chassis.Chassis)
        self.assertEqual(
            '/redfish/v1/Chassis/1U', actual_chassis[0].path)


class ManagerCollectionTestCase(base.TestCase):

    def setUp(self):
        super(ManagerCollectionTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'manager_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        self.managers = manager.ManagerCollection(
            self.conn, '/redfish/v1/Managers', redfish_version='1.0.2')

    @mock.patch.object(manager, 'Manager', autospec=True)
    def test_get_member(self, Manager_mock):
        self.managers.get_member('/redfish/v1/Managers/BMC')
        Manager_mock.assert_called_once_with(
            self.managers._conn, '/redfish/v1/Managers/BMC',
            redfish_version=self.managers.redfish_version, registries=None,
            root=self.managers.root)

    @mock.patch.object(manager, 'Manager', autospec=True)
    def test_get_members(self, Manager_mock):
        members = self.managers.get_members()
        Manager_mock.assert_called_once_with(
            self.managers._conn, '/redfish/v1/Managers/BMC',
            redfish_version=self.managers.redfish_version, registries=None,
            root=self.managers.root)
        self.assertIsInstance(members, list)
        self.assertEqual(1, len(members))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from dateutil import parser

import sushy
from sushy.resources import constants as res_cons
from sushy.resources.system.storage import constants as store_cons
from sushy.resources.system.storage import volume
from sushy import taskmonitor
from sushy.tests.unit import base


class VolumeTestCase(base.TestCase):

    def setUp(self):
        super(VolumeTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/volume.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.stor_volume = volume.Volume(
            self.conn, '/redfish/v1/Systems/437XR1138R2/Storage/1/Volumes/1',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.stor_volume._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.stor_volume.redfish_version)
        self.assertEqual('1', self.stor_volume.identity)
        self.assertEqual('Virtual Disk 1', self.stor_volume.name)
        self.assertEqual(899527000000, self.stor_volume.capacity_bytes)
        self.assertEqual(sushy.VolumeType.MIRRORED,
                         self.stor_volume.volume_type)
        self.assertFalse(self.stor_volume.encrypted)
        identifiers = self.stor_volume.identifiers
        self.assertIsInstance(identifiers, list)
        self.assertEqual(1, len(identifiers))
        identifier = identifiers[0]
        self.assertEqual(sushy.DurableNameFormat.UUID,
                         identifier.durable_name_format)
        self.assertEqual('38f1818b-111e-463a-aa19-fa54f792e468',
                         identifier.durable_name)
        self.assertIsNone(self.stor_volume.block_size_bytes)

    def test_initialize_immediate(self):
        target_uri = '/redfish/v1/Systems/3/Storage/RAIDIntegrated/' \
                     'Volumes/1/Actions/Volume.Initialize'
        self.stor_volume.initialize(
            store_cons.VolumeInitializeType.FAST,
            apply_time=res_cons.ApplyTime.IMMEDIATE)
        self.stor_volume._conn.post.assert_called_once_with(
            target_uri, data={'InitializeType': 'Fast',
                              '@Redfish.OperationApplyTime': 'Immediate'},
            blocking=True, timeout=500)

    def test_initialize_on_reset(self):
        target_uri = '/redfish/v1/Systems/3/Storage/RAIDIntegrated/' \
                     'Volumes/1/Actions/Volume.Initialize'
        self.stor_volume.initialize(
            store_cons.VolumeInitializeType.SLOW,
            apply_time=res_cons.ApplyTime.ON_RESET)
        self.stor_volume._conn.post.assert_called_once_with(
            target_uri, data={'InitializeType': 'Slow',
                              '@Redfish.OperationApplyTime': 'OnReset'},
            blocking=False, timeout=500)

    def test_delete_immediate(self):
        payload = {}
        self.conn.delete.return_value.status_code = 200
        resource = self.stor_volume.delete(
            payload=payload, apply_time=res_cons.ApplyTime.IMMEDIATE)
        self.stor_volume._conn.delete.assert_called_once_with(
            self.stor_volume._path, data=payload, blocking=True, timeout=500)
        self.assertIsNone(resource)

    def test_delete_on_reset(self):
        payload = {}
        self.conn.delete.return_value.status_code = 202
        self.conn.delete.return_value.headers = {
            'Location': '/redfish/v1/taskmon/4608f7e6',
            'Retry-After': '120'
        }
        self.conn.delete.return_value.json.return_value = {'Id': 3,
                                                           'Name': 'Test'}
        task_mon = self.stor_volume.delete(
            payload=payload, apply_time=res_cons.ApplyTime.ON_RESET,
            timeout=250)
        self.stor_volume._conn.delete.assert_called_once_with(
            self.stor_volume._path, data=payload, blocking=False, timeout=250)
        self.assertIsNotNone(task_mon)
        self.assertIsInstance(task_mon, taskmonitor.TaskMonitor)
        self.assertEqual(task_mon.task_monitor_uri,
                         '/redfish/v1/taskmon/4608f7e6')


class VolumeCollectionTestCase(base.TestCase):

    def setUp(self):
        super(VolumeCollectionTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'volume_collection.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.stor_vol_col = volume.VolumeCollection(
            self.conn, '/redfish/v1/Systems/437XR1138R2/Storage/1/Volumes',
            redfish_version='1.0.2')
        self.stor_vol_col.refresh = mock.Mock()

    def test__parse_attributes(self):
        self.stor_vol_col._parse_attributes(self.json_doc)
        self.assertEqual((
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Volumes/1',
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Volumes/2',
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Volumes/3'),
            self.stor_vol_col.members_identities)

    def test_operation_apply_time_support(self):
        support = self.stor_vol_col.operation_apply_time_support
        self.assertIsNotNone(support)
        self.assertEqual(600, support.maintenance_window_duration_in_seconds)
        self.assertEqual(parser.parse('2017-05-03T23:12:37-05:00'),
                         support.maintenance_window_start_time)
        self.assertEqual('/redfish/v1/Systems/437XR1138R2',
                         support._maintenance_window_resource.resource_uri)
        self.assertEqual(['Immediate', 'OnReset', 'AtMaintenanceWindowStart'],
                         support.supported_values)
        self.assertEqual([res_cons.ApplyTime.IMMEDIATE,
                          res_cons.ApplyTime.ON_RESET,
                          res_cons.ApplyTime.AT_MAINTENANCE_WINDOW_START],
                         support.mapped_supported_values)

    @mock.patch.object(volume, 'Volume', autospec=True)
    def test_get_member(self, Volume_mock):
        self.stor_vol_col.get_member(
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Volumes/1')
        Volume_mock.assert_called_once_with(
            self.stor_vol_col._conn,
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Volumes/1',
            redfish_version=self.stor_vol_col.redfish_version, registries=None,
            root=self.stor_vol_col.root)

    @mock.patch.object(volume, 'Volume', autospec=True)
    def test_get_members(self, Volume_mock):
        members = self.stor_vol_col.get_members()
        calls = [
            mock.call(self.stor_vol_col._conn,
                      '/redfish/v1/Systems/437XR1138R2/Storage/1/Volumes/1',
                      redfish_version=self.stor_vol_col.redfish_version,
                      registries=None, root=self.stor_vol_col.root),
            mock.call(self.stor_vol_col._conn,
                      '/redfish/v1/Systems/437XR1138R2/Storage/1/Volumes/2',
                      redfish_version=self.stor_vol_col.redfish_version,
                      registries=None, root=self.stor_vol_col.root),
            mock.call(self.stor_vol_col._conn,
                      '/redfish/v1/Systems/437XR1138R2/Storage/1/Volumes/3',
                      redfish_version=self.stor_vol_col.redfish_version,
                      registries=None, root=self.stor_vol_col.root),
        ]
        Volume_mock.assert_has_calls(calls)
        self.assertIsInstance(members, list)
        self.assertEqual(3, len(members))

    def test_max_size_bytes(self):
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        file_names = ['sushy/tests/unit/json_samples/volume.json',
                      'sushy/tests/unit/json_samples/volume2.json',
                      'sushy/tests/unit/json_samples/volume3.json']
        for file_name in file_names:
            with open(file_name) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual(1073741824000, self.stor_vol_col.max_size_bytes)

        # for any subsequent fetching it gets it from the cached value
        self.conn.get.return_value.json.reset_mock()
        self.assertEqual(1073741824000, self.stor_vol_col.max_size_bytes)
        self.conn.get.return_value.json.assert_not_called()

    def test_max_size_bytes_after_refresh(self):
        self.stor_vol_col.refresh()
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        file_names = ['sushy/tests/unit/json_samples/volume.json',
                      'sushy/tests/unit/json_samples/volume2.json',
                      'sushy/tests/unit/json_samples/volume3.json']
        for file_name in file_names:
            with open(file_name) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual(1073741824000, self.stor_vol_col.max_size_bytes)

    def test_create_immediate(self):
        payload = {
            'Name': 'My Volume 4',
            'VolumeType': 'Mirrored',
            'RAIDType': 'RAID1',
            'CapacityBytes': 107374182400
        }
        expected_payload = dict(payload)
        expected_payload['@Redfish.OperationApplyTime'] = 'Immediate'
        with open('sushy/tests/unit/json_samples/volume4.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        self.conn.post.return_value.status_code = 201
        self.conn.post.return_value.headers.return_value = {
            'Location': '/redfish/v1/Systems/437XR1138R2/Storage/1/Volumes/4'
        }
        new_vol = self.stor_vol_col.create(
            payload, apply_time=res_cons.ApplyTime.IMMEDIATE)
        self.stor_vol_col._conn.post.assert_called_once_with(
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Volumes',
            data=expected_payload, blocking=True, timeout=500)
        self.stor_vol_col.refresh.assert_called_once()
        self.assertIsNotNone(new_vol)
        self.assertEqual('4', new_vol.identity)
        self.assertEqual('My Volume 4', new_vol.name)
        self.assertEqual(107374182400, new_vol.capacity_bytes)
        self.assertEqual(sushy.VolumeType.MIRRORED, new_vol.volume_type)
        self.assertEqual(sushy.RAIDType.RAID1, new_vol.raid_type)

    def test_create_on_reset(self):
        payload = {
            'Name': 'My Volume 4',
            'VolumeType': 'Mirrored',
            'RAIDType': 'RAID1',
            'CapacityBytes': 107374182400
        }
        expected_payload = dict(payload)
        expected_payload['@Redfish.OperationApplyTime'] = 'OnReset'
        with open('sushy/tests/unit/json_samples/task.json') as f:
            self.conn.post.return_value.json.return_value = json.load(f)

        self.conn.post.return_value.content.return_value = "Something"
        self.conn.post.return_value.status_code = 202
        self.conn.post.return_value.headers = {
            'Location': '/redfish/v1/taskmon/4608f7e6',
            'Retry-After': '120'
        }
        task_mon = self.stor_vol_col.create(
            payload, apply_time=res_cons.ApplyTime.ON_RESET)
        self.stor_vol_col._conn.post.assert_called_once_with(
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Volumes',
            data=expected_payload, blocking=False, timeout=500)
        self.assertIsNotNone(task_mon)
        self.assertIsInstance(task_mon, taskmonitor.TaskMonitor)
        self.assertEqual(task_mon.task_monitor_uri,
                         '/redfish/v1/TaskService/Tasks/545')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from unittest import mock

from sushy import exceptions
from sushy.resources.compositionservice import constants as res_block_cons
from sushy.resources.compositionservice import resourceblock
from sushy.resources import constants as res_cons
from sushy.tests.unit import base


class ResourceBlockTestCase(base.TestCase):

    def setUp(self):
        super(ResourceBlockTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/resourceblock.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.res_block = resourceblock.ResourceBlock(
            self.conn,
            '/redfish/v1/CompositionService/ResourceBlocks/DriveBlock3',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.res_block._parse_attributes(self.json_doc)
        self.assertEqual(
            res_block_cons.CompositionState.COMPOSED,
            self.res_block.composition_status.composition_state)
        self.assertEqual(1, self.res_block.composition_status.max_compositions)
        self.assertEqual(
            0, self.res_block.composition_status.number_of_compositions)
        self.assertFalse(self.res_block.composition_status.reserved_state)
        self.assertTrue(self.res_block.composition_status.sharing_capable)
        self.assertFalse(self.res_block.composition_status.sharing_enabled)
        self.assertEqual('ResourceBlock1', self.res_block.description)
        self.assertEqual('DriveBlock3', self.res_block.identity)
        self.assertEqual('Drive Block 3', self.res_block.name)
        self.assertEqual(
            res_block_cons.ResourceBlockType.STORAGE,
            self.res_block.resource_block_type)
        self.assertEqual(
            res_cons.State.ENABLED,
            self.res_block.status.state)
        self.assertEqual(res_cons.Health.OK, self.res_block.status.health)
        exp_path = '/redfish/v1/CompositionService/ResourceBlocks/DriveBlock3'
        self.assertEqual(exp_path, self.res_block.path)

    def test__parse_attributes_missing_identity(self):
        self.res_block.json.pop('Id')
        self.assertRaisesRegex(
            exceptions.MissingAttributeError, 'attribute Id',
            self.res_block._parse_attributes, self.json_doc)


class ResourceBlockCollectionTestCase(base.TestCase):

    def setUp(self):
        super(ResourceBlockCollectionTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'resourceblock_collection.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.res_block_col = resourceblock.ResourceBlockCollection(
            self.conn, '/redfish/v1/CompositionService/ResourceBlocks',
            '1.0.2', None)

    def test__parse_attributes(self):
        path = '/redfish/v1/CompositionService/ResourceBlocks/ComputeBlock1'
        self.res_block_col._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.res_block_col.redfish_version)
        self.assertEqual(
            'Resource Block Collection',
            self.res_block_col.name)
        self.assertEqual((path,), self.res_block_col.members_identities)

    @mock.patch.object(resourceblock, 'ResourceBlock', autospec=True)
    def test_get_member(self, mock_resourceblock):
        path = '/redfish/v1/CompositionService/ResourceBlocks/ComputeBlock1'
        self.res_block_col.get_member(path)
        mock_resourceblock.assert_called_once_with(
            self.res_block_col._conn, path,
            redfish_version=self.res_block_col.redfish_version,
            registries=None, root=self.res_block_col.root)

    @mock.patch.object(resourceblock, 'ResourceBlock', autospec=True)
    def test_get_members(self, mock_resourceblock):
        path = '/redfish/v1/CompositionService/ResourceBlocks/ComputeBlock1'
        members = self.res_block_col.get_members()
        mock_resourceblock.assert_called_once_with(
            self.res_block_col._conn, path,
            redfish_version=self.res_block_col.redfish_version,
            registries=None, root=self.res_block_col.root)
        self.assertIsInstance(members, list)
        self.assertEqual(1, len(members))



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock


from sushy import exceptions
from sushy.resources.sessionservice import session
from sushy.resources.sessionservice import sessionservice
from sushy.tests.unit import base


class SessionServiceTestCase(base.TestCase):

    def setUp(self):
        super(SessionServiceTestCase, self).setUp()
        self.conn = mock.MagicMock()
        with open('sushy/tests/unit/json_samples/session_service.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.sess_serv_inst = sessionservice.SessionService(
            self.conn, '/redfish/v1/SessionService',
            redfish_version='1.0.2')

    def test__init_throws_exception(self):
        self.conn.get.return_value.json.reset_mock()
        self.conn.get.return_value.json.side_effect = (
            exceptions.AccessError(
                'GET', 'any_url', mock.MagicMock()))
        # Previously sushy would just mask these, but now we raise
        # the access error to the user.
        self.assertRaises(exceptions.AccessError,
                          sessionservice.SessionService,
                          self.conn, '/redfish/v1/SessionService',
                          redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.sess_serv_inst._parse_attributes(self.json_doc)
        exp_path = '/redfish/v1/SessionService'
        self.assertEqual('1.0.2', self.sess_serv_inst.redfish_version)
        self.assertEqual('SessionService', self.sess_serv_inst.identity)
        self.assertEqual('Session Service', self.sess_serv_inst.name)
        self.assertEqual(30, self.sess_serv_inst.session_timeout)
        self.assertEqual(exp_path, self.sess_serv_inst.path)

    def test__get_sessions_collection_path(self):
        self.sess_serv_inst.json.pop('Sessions')
        self.assertRaisesRegex(
            exceptions.MissingAttributeError, 'attribute Sessions',
            self.sess_serv_inst._get_sessions_collection_path)

    @mock.patch.object(session, 'SessionCollection', autospec=True)
    def test_session_collection(self, mock_sess_col):
        self.sess_serv_inst.sessions
        mock_sess_col.assert_called_once_with(
            self.sess_serv_inst._conn,
            '/redfish/v1/SessionService/Sessions',
            self.sess_serv_inst.redfish_version, None,
            self.sess_serv_inst.root)

    def test_create_session(self):
        with open('sushy/tests/unit/json_samples/'
                  'session_creation_headers.json') as f:
            self.conn.post.return_value.headers = json.load(f)

        session_key, session_uri = (
            self.sess_serv_inst.create_session('foo', 'secret'))
        self.assertEqual('adc530e2016a0ea98c76c087f0e4b76f', session_key)
        self.assertEqual(
            '/redfish/v1/SessionService/Sessions/151edd65d41c0b89',
            session_uri)

    def test_create_session_unknown_path(self):
        del self.sess_serv_inst.json['Sessions']
        with open('sushy/tests/unit/json_samples/'
                  'session_creation_headers.json') as f:
            self.conn.post.return_value.headers = json.load(f)

        session_key, session_uri = (
            self.sess_serv_inst.create_session('foo', 'secret'))
        self.assertEqual('adc530e2016a0ea98c76c087f0e4b76f', session_key)
        self.assertEqual(
            '/redfish/v1/SessionService/Sessions/151edd65d41c0b89',
            session_uri)
        uri = self.sess_serv_inst.path + '/Sessions'
        data = {'UserName': 'foo', 'Password': 'secret'}
        self.conn.post.assert_called_once_with(uri,
                                               data=data)

    def test_create_session_missing_x_auth_token(self):
        with open('sushy/tests/unit/json_samples/'
                  'session_creation_headers.json') as f:
            self.conn.post.return_value.headers = json.load(f)

        self.conn.post.return_value.headers.pop('X-Auth-Token')
        self.assertRaisesRegex(
            exceptions.MissingXAuthToken, 'No X-Auth-Token returned',
            self.sess_serv_inst.create_session, 'foo', 'bar')

    @mock.patch.object(sessionservice, 'LOG', autospec=True)
    def test_create_session_missing_location(self, mock_LOG):
        with open('sushy/tests/unit/json_samples/'
                  'session_creation_headers.json') as f:
            self.conn.post.return_value.headers = json.load(f)

        self.conn.post.return_value.headers.pop('Location')
        self.sess_serv_inst.create_session('foo', 'bar')
        self.assertTrue(mock_LOG.warning.called)

    def _setUp_sessions(self):
        self.conn.get.return_value.json.reset_mock()
        successive_return_values = []
        with open('sushy/tests/unit/json_samples/session.json') as f:
            successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

    def test_sessions(self):
        # | GIVEN |
        self._setUp_sessions()
        # | WHEN |
        actual_sessions = self.sess_serv_inst.sessions
        # | THEN |
        self.assertIsInstance(actual_sessions, session.SessionCollection)
        self.conn.get.return_value.json.assert_called_once_with()

        # reset mock
        self.conn.get.return_value.json.reset_mock()

        # | WHEN & THEN |
        # tests for same object on invoking subsequently
        self.assertIs(actual_sessions, self.sess_serv_inst.sessions)
        self.conn.get.return_value.json.assert_not_called()

    def test_sessions_on_refresh(self):
        # | GIVEN |
        self._setUp_sessions()
        # | WHEN & THEN |
        self.assertIsInstance(self.sess_serv_inst.sessions,
                              session.SessionCollection)

        self.conn.get.return_value.json.side_effect = None
        # On refreshing the sess_serv_inst instance...
        with open('sushy/tests/unit/json_samples/session.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        self.sess_serv_inst.refresh(force=True)

        # | WHEN & THEN |
        self.assertFalse(self.sess_serv_inst.sessions._is_stale)

    def test_close_session(self):
        self.sess_serv_inst.close_session('session/identity')
        self.conn.delete.assert_called_once_with('session/identity')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from sushy.resources import constants as res_cons
from sushy.resources.system.network import constants as net_cons
from sushy.resources.system.network import device_function
from sushy.tests.unit import base


class NetworkDeviceFunctionTestCase(base.TestCase):

    def setUp(self):
        super(NetworkDeviceFunctionTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'network_device_function.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.function = device_function.NetworkDeviceFunction(
            self.conn,
            '/redfish/v1/Chassis/Blade1/NetworkAdapters/'
            'NIC.Integrated.1/NetworkDeviceFunctions/NIC.Integrated.1-2-1',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.function._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.function.redfish_version)
        self.assertEqual('NIC.Integrated.1-2-1', self.function.identity)
        self.assertEqual('NetworkDeviceFunction', self.function.name)
        self.assertEqual('NetworkDeviceFunction', self.function.description)
        self.assertEqual(res_cons.State.ENABLED, self.function.status.state)
        self.assertEqual(res_cons.Health.OK, self.function.status.health)
        self.assertEqual(res_cons.Health.OK,
                         self.function.status.health_rollup)
        self.assertEqual(self.function.type,
                         net_cons.NetworkDeviceTechnology.ETHERNET)
        self.assertEqual([
            net_cons.NetworkDeviceTechnology.DISABLED,
            net_cons.NetworkDeviceTechnology.ETHERNET,
            net_cons.NetworkDeviceTechnology.iSCSI,
        ], self.function.capabilities)
        self.assertEqual(1,
                         len(self.function.fibre_channel.boot_targets))
        self.assertEqual('ABCEDFG',
                         self.function.fibre_channel.boot_targets[0].wwpn)
        self.assertEqual('1234',
                         self.function.fibre_channel.boot_targets[0].lun_id)
        self.assertIsNone(
            self.function.fibre_channel.boot_targets[0].priority
        )
        self.assertEqual(127, self.function.max_virtual_functions)
        self.assertEqual('01:01:01:01:01:D1',
                         self.function.ethernet.mac_address)
        self.assertEqual('01:01:01:01:01:D1',
                         self.function.ethernet.permanent_mac_address)
        self.assertEqual(1234,
                         self.function.ethernet.mtu_size)
        self.assertEqual(1,
                         self.function.ethernet.vlan.vlan_id)
        self.assertTrue(self.function.ethernet.vlan.vlan_enabled)
        self.assertEqual(net_cons.NetworkAuthenticationMethod.NONE,
                         self.function.iscsi_boot.authentication_method)
        self.assertEqual(net_cons.IPAddressType.IPV4,
                         self.function.iscsi_boot.ip_address_type)
        self.assertEqual('0.0.0.0',
                         self.function.iscsi_boot.initiator_ip_address)
        self.assertEqual('0.0.0.0',
                         self.function.iscsi_boot.initiator_default_gateway)
        self.assertEqual('0.0.0.0',
                         self.function.iscsi_boot.initiator_netmask)
        self.assertEqual('0.0.0.0',
                         self.function.iscsi_boot.primary_dns)
        self.assertEqual('0.0.0.0',
                         self.function.iscsi_boot.primary_dns)
        self.assertEqual(0, self.function.iscsi_boot.primary_lun)
        self.assertEqual('0.0.0.0',
                         self.function.iscsi_boot.primary_target_ip_address)
        self.assertEqual(3260,
                         self.function.iscsi_boot.primary_target_tcp_port)
        self.assertFalse(self.function.iscsi_boot.secondary_vlan_enabled)
        self.assertEqual(1,
                         self.function.iscsi_boot.secondary_vlan_id)
        self.assertEqual(123,
                         self.function.iscsi_boot.secondary_target_tcp_port)


class NetworkDeviceFunctionCollectionTestCase(base.TestCase):

    def setUp(self):
        super(NetworkDeviceFunctionCollectionTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'network_device_function_collection.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.function_col = device_function.NetworkDeviceFunctionCollection(
            self.conn, '/redfish/v1/Chassis/System.Embedded.1/NetworkAdapters/'
                       'NIC.Integrated.1/NetworkDeviceFunctions',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.function_col._parse_attributes(self.json_doc)
        self.assertEqual((
            '/redfish/v1/Chassis/System.Embedded.1/NetworkAdapters/'
            'NIC.Integrated.1/NetworkDeviceFunctions/NIC.Integrated.1-1-1',
            '/redfish/v1/Chassis/System.Embedded.1/NetworkAdapters/'
            'NIC.Integrated.1/NetworkDeviceFunctions/NIC.Integrated.1-2-1'
        ),
            self.function_col.members_identities)

    @mock.patch.object(device_function, 'NetworkDeviceFunction', autospec=True)
    def test_get_member(self, device_function_mock):
        self.function_col.get_member(
            '/redfish/v1/Chassis/System.Embedded.1/NetworkAdapters/'
            'NIC.Integrated.1/NetworkDeviceFunctions/NIC.Integrated.1-1-1',
        )
        device_function_mock.assert_called_once_with(
            self.function_col._conn,
            '/redfish/v1/Chassis/System.Embedded.1/NetworkAdapters/'
            'NIC.Integrated.1/NetworkDeviceFunctions/NIC.Integrated.1-1-1',
            redfish_version=self.function_col.redfish_version,
            registries=None,
            root=self.function_col.root)

    @mock.patch.object(device_function, 'NetworkDeviceFunction', autospec=True)
    def test_get_members(self, device_function_mock):
        members = self.function_col.get_members()
        device_function_mock.assert_has_calls([
            mock.call(
                self.function_col._conn,
                '/redfish/v1/Chassis/System.Embedded.1/NetworkAdapters/'
                'NIC.Integrated.1/NetworkDeviceFunctions/NIC.Integrated.1-1-1',
                redfish_version=self.function_col.redfish_version,
                registries=None,
                root=self.function_col.root
            ),
            mock.call(
                self.function_col._conn,
                '/redfish/v1/Chassis/System.Embedded.1/NetworkAdapters/'
                'NIC.Integrated.1/NetworkDeviceFunctions/NIC.Integrated.1-2-1',
                redfish_version=self.function_col.redfish_version,
                registries=None,
                root=self.function_col.root
            ),
        ])
        self.assertIsInstance(members, list)
        self.assertEqual(2, len(members))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from unittest import mock

from sushy import exceptions
from sushy.resources.compositionservice import resourcezone
from sushy.resources import constants as res_cons
from sushy.tests.unit import base


class ResourceZoneTestCase(base.TestCase):

    def setUp(self):
        super(ResourceZoneTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/resourcezone.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.res_zone = resourcezone.ResourceZone(
            self.conn,
            '/redfish/v1/CompositionService/ResourceZones/1',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.res_zone._parse_attributes(self.json_doc)
        self.assertEqual('ResourceZone1', self.res_zone.description)
        self.assertEqual('1', self.res_zone.identity)
        self.assertEqual('Resource Zone 1', self.res_zone.name)
        self.assertEqual(
            res_cons.State.ENABLED,
            self.res_zone.status.state)
        self.assertEqual(
            res_cons.Health.OK,
            self.res_zone.status.health)
        exp_path = '/redfish/v1/CompositionService/ResourceZones/1'
        self.assertEqual(exp_path, self.res_zone.path)

    def test__parse_attributes_missing_identity(self):
        self.res_zone.json.pop('Id')
        self.assertRaisesRegex(
            exceptions.MissingAttributeError, 'attribute Id',
            self.res_zone._parse_attributes, self.json_doc)


class ResourceZoneCollectionTestCase(base.TestCase):

    def setUp(self):
        super(ResourceZoneCollectionTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'resourcezone_collection.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.res_zone_col = resourcezone.ResourceZoneCollection(
            self.conn, '/redfish/v1/CompositionService/ResourceZones',
            '1.0.2', None)

    def test__parse_attributes(self):
        path = '/redfish/v1/CompositionService/ResourceZones/1'
        self.res_zone_col._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.res_zone_col.redfish_version)
        self.assertEqual('Resource Zone Collection', self.res_zone_col.name)
        self.assertEqual((path,), self.res_zone_col.members_identities)

    @mock.patch.object(resourcezone, 'ResourceZone', autospec=True)
    def test_get_member(self, mock_resourcezone):
        path = '/redfish/v1/CompositionService/ResourceZones/1'
        self.res_zone_col.get_member(path)
        mock_resourcezone.assert_called_once_with(
            self.res_zone_col._conn, path,
            redfish_version=self.res_zone_col.redfish_version, registries=None,
            root=self.res_zone_col.root)

    @mock.patch.object(resourcezone, 'ResourceZone', autospec=True)
    def test_get_members(self, mock_resourcezone):
        path = '/redfish/v1/CompositionService/ResourceZones/1'
        members = self.res_zone_col.get_members()
        mock_resourcezone.assert_called_once_with(
            self.res_zone_col._conn, path,
            redfish_version=self.res_zone_col.redfish_version, registries=None,
            root=self.res_zone_col.root)
        self.assertIsInstance(members, list)
        self.assertEqual(1, len(members))



#    Copyright (c) 2021 Anexia Internetdienstleistungs GmbH
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/v1/NetworkAdapter.v1_3_0.json

from sushy.resources import base
from sushy.resources import common
from sushy.resources.system.network import device_function
from sushy.resources.system.network import port
from sushy import utils


class NetworkAdapter(base.ResourceBase):
    description = base.Field('Description')
    """Human-readable description of the resource"""

    identity = base.Field('Id', required=True)
    """The network adapter identity string"""

    manufacturer = base.Field("Manufacturer")
    """The manufacturer of this network adapter"""

    model = base.Field("Model")
    """The model of this network adapter"""

    name = base.Field('Name')
    """The name of the network adapter"""

    part_number = base.Field("PartNumber")
    """The part number of the network adapter"""

    serial_number = base.Field("SerialNumber")
    """The serial number of the network adapter"""

    status = common.StatusField("Status")
    """The status"""

    @property
    @utils.cache_it
    def network_device_functions(self):
        """Property to reference `NetworkDeviceFunctionCollection` instance

        It is set once when the first time it is queried. On refresh,
        this property is marked as stale (greedy-refresh not done).
        Here the actual refresh of the sub-resource happens, if stale.
        """

        return device_function.NetworkDeviceFunctionCollection(
            self._conn,
            path=utils.get_sub_resource_path_by(
                self,
                "NetworkDeviceFunctions"
            ),
            redfish_version=self.redfish_version,
            registries=self.registries,
            root=self.root
        )

    @property
    @utils.cache_it
    def network_ports(self):
        """Property to reference `NetworkPortCollection` instance

        It is set once when the first time it is queried. On refresh,
        this property is marked as stale (greedy-refresh not done).
        Here the actual refresh of the sub-resource happens, if stale.
        """

        return port.NetworkPortCollection(
            self._conn,
            path=utils.get_sub_resource_path_by(self, "NetworkPorts"),
            redfish_version=self.redfish_version,
            registries=self.registries,
            root=self.root
        )


class NetworkAdapterCollection(base.ResourceCollectionBase):

    @property
    def _resource_type(self):
        return NetworkAdapter



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This is referred from Redfish standard schema.
# http://redfish.dmtf.org/schemas/v1/SecureBoot.v1_1_0.json

import logging

from sushy import exceptions
from sushy.resources import base
from sushy.resources import common
from sushy.resources.system import constants
from sushy.resources.system import secure_boot_database
from sushy import utils

LOG = logging.getLogger(__name__)


class ResetKeysActionField(common.ActionField):

    allowed_values = base.Field('ResetKeysType@Redfish.AllowableValues',
                                adapter=list)


class ActionsField(base.CompositeField):

    reset_keys = ResetKeysActionField('#SecureBoot.ResetKeys')
    """Action that resets the UEFI Secure Boot keys."""


class SecureBoot(base.ResourceBase):

    identity = base.Field('Id', required=True)
    """The Bios resource identity string"""

    name = base.Field('Name')
    """The name of the resource"""

    description = base.Field('Description')
    """Human-readable description of the BIOS resource"""

    current_boot = base.MappedField('SecureBootCurrentBoot',
                                    constants.SecureBootCurrentBoot)
    """The UEFI Secure Boot state during the current boot cycle."""

    enabled = base.Field('SecureBootEnable')
    """Whether the UEFI Secure Boot takes effect on next boot.

    This property can be enabled in UEFI boot mode only.
    """

    mode = base.MappedField('SecureBootMode', constants.SecureBootMode)
    """The current UEFI Secure Boot Mode."""

    # TODO(dtantsur): SecureBootDatabases

    _actions = ActionsField('Actions')

    def __init__(self, connector, path, redfish_version=None, registries=None,
                 root=None):
        """A class representing secure boot settings.

        :param connector: A Connector instance
        :param path: Sub-URI path to the SecureBoot resource
        :param registries: Dict of message registries to be used when
            parsing messages of attribute update status
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        super().__init__(connector, path, redfish_version=redfish_version,
                         registries=registries, root=root)

    @property
    @utils.cache_it
    def databases(self):
        """A collection of secure boot databases.

        It is set once when the first time it is queried. On refresh,
        this property is marked as stale (greedy-refresh not done).
        Here the actual refresh of the sub-resource happens, if stale.

        :raises: MissingAttributeError if 'SecureBootDatabases/@odata.id' field
            is missing.
        :returns: `SimpleStorageCollection` instance
        """
        return secure_boot_database.SecureBootDatabaseCollection(
            self._conn, utils.get_sub_resource_path_by(
                self, "SecureBootDatabases"),
            redfish_version=self.redfish_version,
            registries=self.registries, root=self.root)

    def _get_reset_action_element(self):
        reset_action = self._actions.reset_keys
        if not reset_action:
            raise exceptions.MissingActionError(action='#SecureBoot.ResetKeys',
                                                resource=self._path)
        return reset_action

    def get_allowed_reset_keys_values(self):
        """Get the allowed values for resetting the keys.

        :returns: A set with the allowed values.
        """
        reset_action = self._get_reset_action_element()

        if not reset_action.allowed_values:
            LOG.warning('Could not figure out the allowed values for the '
                        'reset keys action for %s', self.identity)
            return set(constants.SecureBootResetKeysType)

        return {v for v in constants.SecureBootResetKeysType
                if v.value in reset_action.allowed_values}

    def reset_keys(self, reset_type):
        """Reset secure boot keys.

        :param reset_type: Reset type, one of `SECURE_BOOT_RESET_KEYS_*`
            constants.
        """
        valid_resets = self.get_allowed_reset_keys_values()
        if reset_type not in valid_resets:
            raise exceptions.InvalidParameterValueError(
                parameter='reset_type', value=reset_type,
                valid_values=valid_resets)

        reset_type = constants.SecureBootResetKeysType(reset_type).value
        target_uri = self._get_reset_action_element().target_uri
        self._conn.post(target_uri, data={'ResetKeysType': reset_type})

    def set_enabled(self, enabled):
        """Enable/disable secure boot.

        :param enabled: True, if secure boot is enabled for next boot.
        """
        if not isinstance(enabled, bool):
            raise exceptions.InvalidParameterValueError(
                "Expected a boolean for 'enabled', got %r" % enabled)

        self._conn.patch(self.path, data={'SecureBootEnable': enabled})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/Settings.v1_2_0.json

import logging

from dateutil import parser

from sushy.resources import base
from sushy.resources import common
from sushy.resources import constants as res_cons
from sushy.resources.registry import message_registry

# Settings update statuses

UPDATE_UNKNOWN = 0
"""Update status unknown"""

UPDATE_SUCCESS = 1
"""Update was successful"""

UPDATE_FAILURE = 2
"""Update encountered errors"""

UPDATE_PENDING = 3
"""Update waiting for being applied"""

NO_UPDATES = 4
"""No updates made"""


class SettingsUpdate(object):
    """Contains Settings update status and details of the update"""

    def __init__(self, status, messages):
        self._status = status
        self._messages = messages

    @property
    def status(self):
        """The status of the update"""
        return self._status

    @property
    def messages(self):
        """List of :class:`.MessageListField` with messages from the update"""
        return self._messages


LOG = logging.getLogger(__name__)


class MaintenanceWindowField(base.CompositeField):

    maintenance_window_duration_in_seconds = base.Field(
        'MaintenanceWindowDurationInSeconds',
        required=True)
    """The expiry time of maintenance window in seconds"""

    maintenance_window_start_time = base.Field(
        'MaintenanceWindowStartTime',
        required=True,
        adapter=parser.parse)
    """The start time of a maintenance window"""


class SettingsApplyTimeField(base.CompositeField):
    def __init__(self):
        super(SettingsApplyTimeField, self).__init__(
            path="@Redfish.SettingsApplyTime")

    apply_time = base.Field('ApplyTime', adapter=str)
    """When the future configuration should be applied"""

    apply_time_allowable_values = base.Field(
        'ApplyTime@Redfish.AllowableValues', adapter=list)
    """The list of allowable ApplyTime values"""

    maintenance_window_start_time = base.Field('MaintenanceWindowStartTime',
                                               adapter=parser.parse)
    """The start time of a maintenance window"""

    maintenance_window_duration_in_seconds = base.Field(
        'MaintenanceWindowDurationInSeconds', adapter=int)
    """The expiry time of maintenance window in seconds"""


class SettingsField(base.CompositeField):
    """The settings of a resource

    Represents the future state and configuration of the resource. The
    field is added to resources that support future state and
    configuration.

    This field includes several properties to help clients monitor when
    the resource is consumed by the service and determine the results of
    applying the values, which may or may not have been successful.
    """

    def __init__(self):
        super(SettingsField, self).__init__(path="@Redfish.Settings")

    time = base.Field('Time')
    """Indicates the time the settings were applied to the server"""

    _etag = base.Field('ETag')
    """The ETag of the resource to which the settings were applied,
    after the application
    """

    _settings_object_idref = common.IdRefField("SettingsObject")
    """Reference to the resource the client may PUT/PATCH in order
    to change this resource
    """

    _supported_apply_times = base.MappedListField(
        'SupportedApplyTimes',
        res_cons.ApplyTime)
    """List of supported apply times"""

    @property
    def maintenance_window(self):
        """MaintenanceWindow field

        Indicates if a given resource has a maintenance window assignment
        for applying settings or operations
        """
        LOG.warning('The @Redfish.MaintenanceWindow annotation does not '
                    'appear within @Redfish.Settings. Instead use the '
                    'maintenance_window property in the target resource '
                    '(e.g. System resource)')
        return None

    messages = base.MessageListField("Messages")
    """Represents the results of the last time the values of the Settings
    resource were applied to the server"""

    @property
    def operation_apply_time_support(self):
        """OperationApplyTimeSupport field

        Indicates if a client is allowed to request for a specific apply
        time of a create, delete, or action operation of a given resource
        """
        LOG.warning('Redfish ApplyTime annotations do not appear within '
                    '@Redfish.Settings. Instead use the apply_time_settings '
                    'property in the target resource (e.g. Bios resource)')
        return None

    def commit(self, connector, value):
        """Commits new settings values

        The new values will be applied when the system or a service
        restarts.

        :param connector: A Connector instance
        :param value: Value representing JSON whose structure is specific
            to each resource and the caller must format it correctly
        :returns: Response object
        """

        return connector.patch(self.resource_uri, data=value)

    @property
    def resource_uri(self):
        return self._settings_object_idref.resource_uri

    def get_status(self, registries):
        """Determines the status of last update based

        Uses message id-s and severity to determine the status.

        :param registries: registries to use to parse message
        :returns: :class:`.SettingsUpdate` object containing status
            and any messages
        """

        if not self.time:
            return SettingsUpdate(NO_UPDATES, None)

        parsed_msgs = []
        for m in self.messages:
            parsed_msgs.append(
                message_registry.parse_message(registries, m))
        any_errors = any(m for m in parsed_msgs
                         if m.severity != res_cons.Severity.OK)

        if any_errors:
            status = UPDATE_FAILURE
        else:
            status = UPDATE_SUCCESS
        return SettingsUpdate(status, parsed_msgs)
